@@ -24,6 +24,7 @@
 #include <sys/types.h>
 #include <unistd.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #include <netdb.h> 
 #include <netinet/in.h>
 #include <X11/Xauth.h>
@@ -37,7 +38,7 @@
 #include "auth.h"
 
 /* Local prototypes */
-static void gdm_auth_purge (GdmDisplay *d, FILE *af);
+static FILE *gdm_auth_purge (GdmDisplay *d, FILE *af, gboolean remove_when_empty);
 
 /* Configuration option variables */
 extern gchar *GdmServAuthDir;
@@ -70,31 +71,37 @@ display_add_error (GdmDisplay *d)
 }
 
 static gboolean
-add_auth_entry (GdmDisplay *d, FILE *af, FILE *af2,
+add_auth_entry (GdmDisplay *d, 
+		GSList **authlist,
+		FILE *af, FILE *af2,
 		unsigned short family, const char *addr, int addrlen)
 {
 	Xauth *xa;
 	gchar *dispnum;
 
-	if (!d)
+	if G_UNLIKELY (!d)
 		return FALSE;
 
 	xa = malloc (sizeof (Xauth));
 
-	if (xa == NULL)
+	if G_UNLIKELY (xa == NULL)
 		return FALSE;
 
 	xa->family = family;
-	xa->address = malloc (addrlen);
-	if (xa->address == NULL) {
-		free (xa);
-		return FALSE;
+	if (addrlen == 0) {
+		xa->address = NULL;
+		xa->address_length = 0;
+	} else {
+		xa->address = malloc (addrlen);
+		if G_UNLIKELY (xa->address == NULL) {
+			free (xa);
+			return FALSE;
+		}
+
+		memcpy (xa->address, addr, addrlen);
+		xa->address_length = addrlen;
 	}
 
-
-	memcpy (xa->address, addr, addrlen);
-	xa->address_length = addrlen;
-
 	dispnum = g_strdup_printf ("%d", d->dispnum);
 	xa->number = strdup (dispnum);
 	xa->number_length = strlen (dispnum);
@@ -103,7 +110,7 @@ add_auth_entry (GdmDisplay *d, FILE *af, FILE *af2,
 	xa->name = strdup ("MIT-MAGIC-COOKIE-1");
 	xa->name_length = strlen ("MIT-MAGIC-COOKIE-1");
 	xa->data = malloc (16);
-	if (xa->data == NULL) {
+	if G_UNLIKELY (xa->data == NULL) {
 		free (xa->number);
 		free (xa->name);
 		free (xa->address);
@@ -113,20 +120,9 @@ add_auth_entry (GdmDisplay *d, FILE *af, FILE *af2,
 	memcpy (xa->data, d->bcookie, 16);
 	xa->data_length = 16;
 
-	errno = 0;
-	if ( ! XauWriteAuth (af, xa)) {
-		free (xa->data);
-		free (xa->number);
-		free (xa->name);
-		free (xa->address);
-		free (xa);
-		display_add_error (d);
-		return FALSE;
-	}
-
-	if (af2 != NULL) {
+	if (af != NULL) {
 		errno = 0;
-		if ( ! XauWriteAuth (af2, xa)) {
+		if G_UNLIKELY ( ! XauWriteAuth (af, xa)) {
 			free (xa->data);
 			free (xa->number);
 			free (xa->name);
@@ -135,9 +131,22 @@ add_auth_entry (GdmDisplay *d, FILE *af, FILE *af2,
 			display_add_error (d);
 			return FALSE;
 		}
+
+		if (af2 != NULL) {
+			errno = 0;
+			if G_UNLIKELY ( ! XauWriteAuth (af2, xa)) {
+				free (xa->data);
+				free (xa->number);
+				free (xa->name);
+				free (xa->address);
+				free (xa);
+				display_add_error (d);
+				return FALSE;
+			}
+		}
 	}
 
-	d->auths = g_slist_append (d->auths, xa);
+	*authlist = g_slist_append (*authlist, xa);
 
 	return TRUE;
 }
@@ -155,12 +164,8 @@ gboolean
 gdm_auth_secure_display (GdmDisplay *d)
 {
     FILE *af, *af_gdm;
-    gboolean is_local = FALSE;
-    const char lo[] = {127,0,0,1};
-    guint i;
-    const GList *local_addys = NULL;
 
-    if (!d)
+    if G_UNLIKELY (!d)
 	return FALSE;
 
     umask (022);
@@ -184,7 +189,7 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    authfd = g_mkstemp (d->authfile);
 	    umask (022);
 
-	    if (authfd == -1) {
+	    if G_UNLIKELY (authfd == -1) {
 		    gdm_error (_("%s: Could not make new cookie file in %s"),
 			       "gdm_auth_secure_display", GdmUserAuthFB);
 		    g_free (d->authfile);
@@ -197,7 +202,7 @@ gdm_auth_secure_display (GdmDisplay *d)
 
 	    af = fdopen (authfd, "w");
 
-	    if (af == NULL) {
+	    if G_UNLIKELY (af == NULL) {
 		    g_free (d->authfile);
 		    d->authfile = NULL;
 		    return FALSE;
@@ -205,12 +210,12 @@ gdm_auth_secure_display (GdmDisplay *d)
 
 	    /* Make another authfile since the greeter can't read the server/user
 	     * readable file */
-	    d->authfile_gdm = g_strconcat (GdmServAuthDir, "/", d->name, ".Xauth", NULL);
+	    d->authfile_gdm = gdm_make_filename (GdmServAuthDir, d->name, ".Xauth");
 	    af_gdm = gdm_safe_fopen_w (d->authfile_gdm);
 
-	    if (af_gdm == NULL) {
+	    if G_UNLIKELY (af_gdm == NULL) {
 		    gdm_error (_("%s: Cannot safely open %s"),
-			       "gdm_auth_user_remove",
+			       "gdm_auth_secure_display",
 			       d->authfile_gdm);
 
 		    g_free (d->authfile_gdm);
@@ -222,12 +227,12 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    }
     } else {
 	    /* gdm and xserver authfile can be the same, server will run as root */
-	    d->authfile = g_strconcat (GdmServAuthDir, "/", d->name, ".Xauth", NULL);
+	    d->authfile = gdm_make_filename (GdmServAuthDir, d->name, ".Xauth");
 	    af = gdm_safe_fopen_w (d->authfile);
 
-	    if (af == NULL) {
+	    if G_UNLIKELY (af == NULL) {
 		    gdm_error (_("%s: Cannot safely open %s"),
-			       "gdm_auth_user_remove",
+			       "gdm_auth_secure_display",
 			       d->authfile);
 
 		    g_free (d->authfile);
@@ -242,14 +247,7 @@ gdm_auth_secure_display (GdmDisplay *d)
      * have to eat up old authentication cookies before baking new
      * ones... */
     if (SERVER_IS_LOCAL (d) && d->auths) {
-	    GSList *li;
-	   
-	    for (li = d->auths; li != NULL; li = li->next) {
-		    XauDisposeAuth ((Xauth *) li->data);
-		    li->data = NULL;
-	    }
-
-	    g_slist_free (d->auths);
+	    gdm_auth_free_auth_list (d->auths);
 	    d->auths = NULL;
 
 	    g_free (d->cookie);
@@ -266,11 +264,60 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    char hostname[1024];
 
 	    hostname[1023] = '\0';
-	    if (gethostname (hostname, 1023) == 0) {
+	    if G_LIKELY (gethostname (hostname, 1023) == 0) {
 		    g_free (d->hostname);
 		    d->hostname = g_strdup (hostname);
 	    }
-	    local_addys = gdm_peek_local_address_list ();
+    }
+
+    if ( ! add_auth_entry (d, &(d->auths), af, af_gdm, FamilyWild, NULL, 0))
+	    return FALSE;
+
+    gdm_debug ("gdm_auth_secure_display: Setting up access");
+
+    if G_UNLIKELY (fclose (af) < 0) {
+	    display_add_error (d);
+	    return FALSE;
+    }
+    if (af_gdm != NULL) {
+	    if G_UNLIKELY (fclose (af_gdm) < 0) {
+		    display_add_error (d);
+		    return FALSE;
+	    }
+    }
+    ve_setenv ("XAUTHORITY", GDM_AUTHFILE (d), TRUE);
+
+    if G_UNLIKELY (GdmDebug)
+	    gdm_debug ("gdm_auth_secure_display: Setting up access for %s - %d entries", 
+		       d->name, g_slist_length (d->auths));
+
+    return TRUE;
+}
+
+static GSList *
+get_local_auths (GdmDisplay *d)
+{
+    gboolean is_local = FALSE;
+    const char lo[] = {127,0,0,1};
+    guint i;
+    const GList *local_addys = NULL;
+    gboolean added_lo = FALSE;
+    GSList *auths = NULL;
+
+    if G_UNLIKELY (!d)
+	return NULL;
+
+    if (SERVER_IS_LOCAL (d)) {
+	    char hostname[1024];
+
+	    /* reget local host if local as it may have changed */
+	    hostname[1023] = '\0';
+	    if G_LIKELY (gethostname (hostname, 1023) == 0) {
+		    g_free (d->hostname);
+		    d->hostname = g_strdup (hostname);
+	    }
+	    if ( ! d->tcp_disallowed)
+		    local_addys = gdm_peek_local_address_list ();
 
 	    is_local = TRUE;
     } else  {
@@ -287,11 +334,11 @@ gdm_auth_secure_display (GdmDisplay *d)
 
     /* Local access also in case the host is very local */
     if (is_local) {
-	    gdm_debug ("gdm_auth_secure_display: Setting up socket access");
+	    gdm_debug ("get_local_auths: Setting up socket access");
 
-	    if ( ! add_auth_entry (d, af, af_gdm, FamilyLocal,
+	    if ( ! add_auth_entry (d, &auths, NULL, NULL, FamilyLocal,
 				   d->hostname, strlen (d->hostname)))
-		    return FALSE;
+		    goto get_local_auth_error;
 
 	    /* local machine but not local if you get my meaning, add
 	     * the host gotten by gethostname as well if it's different
@@ -302,33 +349,36 @@ gdm_auth_secure_display (GdmDisplay *d)
 		    hostname[1023] = '\0';
 		    if (gethostname (hostname, 1023) == 0 &&
 			strcmp (hostname, d->hostname) != 0) {
-			    if ( ! add_auth_entry (d, af, af_gdm, FamilyLocal,
+			    if ( ! add_auth_entry (d, &auths, NULL, NULL, FamilyLocal,
 						   hostname,
 						   strlen (hostname)))
-				    return FALSE;
+				    goto get_local_auth_error;
 		    }
 	    } else {
 		    /* local machine, perhaps we haven't added
 		     * localhost.localdomain to socket access */
 		    const char *localhost = "localhost.localdomain";
 		    if (strcmp (localhost, d->hostname) != 0) {
-			    if ( ! add_auth_entry (d, af, af_gdm, FamilyLocal,
+			    if ( ! add_auth_entry (d, &auths, NULL, NULL, FamilyLocal,
 						   localhost,
 						   strlen (localhost))) {
-				    return FALSE;
+				    goto get_local_auth_error;
 			    }
 		    }
 	    }
     }
 
-    gdm_debug ("gdm_auth_secure_display: Setting up network access");
+    gdm_debug ("get_local_auths: Setting up network access");
 
     if ( ! SERVER_IS_LOCAL (d)) {
 	    /* we should write out an entry for d->addr since
 	       possibly it is not in d->addrs */
-	    if ( ! add_auth_entry (d, af, af_gdm, FamilyInternet,
+	    if ( ! add_auth_entry (d, &auths, NULL, NULL, FamilyInternet,
 				   (char *)&(d->addr), 4))
-		    return FALSE;
+		    goto get_local_auth_error;
+
+	    if (gdm_is_loopback_addr (&(d->addr)))
+		    added_lo = TRUE;
     }
     
     /* Network access: Write out an authentication entry for each of
@@ -339,9 +389,12 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    if (memcmp (ia, &(d->addr), sizeof (struct in_addr)) == 0)
 		    continue;
 
-	    if ( ! add_auth_entry (d, af, af_gdm, FamilyInternet,
+	    if ( ! add_auth_entry (d, &auths, NULL, NULL, FamilyInternet,
 				   (char *)&ia->s_addr, 4))
-		    return FALSE;
+		    goto get_local_auth_error;
+
+	    if (gdm_is_loopback_addr (ia))
+		    added_lo = TRUE;
     }
 
     /* Network access: Write out an authentication entry for each of
@@ -352,36 +405,42 @@ gdm_auth_secure_display (GdmDisplay *d)
 	    if (ia == NULL)
 		    break;
 
-	    if ( ! add_auth_entry (d, af, af_gdm, FamilyInternet,
+	    if ( ! add_auth_entry (d, &auths, NULL, NULL, FamilyInternet,
 				   (char *)&ia->s_addr, 4))
-		    return FALSE;
+		    goto get_local_auth_error;
+
+	    if (gdm_is_loopback_addr (ia))
+		    added_lo = TRUE;
     }
 
     /* if local server add loopback */
-    if (SERVER_IS_LOCAL (d)) {
-	    if ( ! add_auth_entry (d, af, af_gdm, FamilyInternet,
+    if (SERVER_IS_LOCAL (d) && ! added_lo && ! d->tcp_disallowed) {
+	    if ( ! add_auth_entry (d, &auths, NULL, NULL, FamilyInternet,
 				   lo, 4))
-		    return FALSE;
+		    goto get_local_auth_error;
     }
 
-    fclose (af);
-    if (af_gdm != NULL)
-	    fclose (af_gdm);
-    ve_setenv ("XAUTHORITY", GDM_AUTHFILE (d), TRUE);
+    if G_UNLIKELY (GdmDebug)
+	    gdm_debug ("get_local_auths: Setting up access for %s - %d entries", 
+		       d->name, g_slist_length (auths));
 
-    if (GdmDebug)
-	    gdm_debug ("gdm_auth_secure_display: Setting up access for %s - %d entries", 
-		       d->name, g_slist_length (d->auths));
+    return auths;
 
-    return TRUE;
+get_local_auth_error:
+
+    gdm_auth_free_auth_list (auths);
+
+    return NULL;
 }
 
+
 static gboolean
 try_open_append (const char *file)
 {
 	FILE *fp;
+
 	fp = fopen (file, "a+");
-	if (fp != NULL) {
+	if G_LIKELY (fp != NULL) {
 		fclose (fp);
 		return TRUE;
 	} else {
@@ -389,6 +448,28 @@ try_open_append (const char *file)
 	}
 }
 
+static gboolean
+try_open_read_as_root (const char *file)
+{
+	int fd;
+	uid_t oldeuid = geteuid ();
+	uid_t oldegid = getegid ();
+	setegid (0);
+	seteuid (0);
+
+	fd = open (file, O_RDONLY);
+	if G_UNLIKELY (fd < 0) {
+		seteuid (oldeuid);
+		setegid (oldegid);
+		return FALSE;
+	} else {
+		close (fd);
+		seteuid (oldeuid);
+		setegid (oldegid);
+		return TRUE;
+	}
+}
+
 /**
  * gdm_auth_user_add:
  * @d: Pointer to a GdmDisplay struct
@@ -410,6 +491,7 @@ gdm_auth_user_add (GdmDisplay *d, uid_t user, const char *homedir)
     GSList *auths = NULL;
     gboolean ret = TRUE;
     gboolean automatic_tmp_dir = FALSE;
+    gboolean authdir_is_tmp_dir = FALSE;
     gboolean locked;
 
     if (!d)
@@ -421,14 +503,28 @@ gdm_auth_user_add (GdmDisplay *d, uid_t user, const char *homedir)
     if ( ! ve_string_empty (GdmUserAuthDir) &&
 	strcmp (GdmUserAuthDir, "~") != 0) {
 	    if (strncmp (GdmUserAuthDir, "~/", 2) == 0) {
-		    authdir = g_strconcat (homedir, &GdmUserAuthDir[1], NULL);
+		    authdir = g_build_filename (homedir, &GdmUserAuthDir[2], NULL);
 	    } else {
 		    authdir = g_strdup (GdmUserAuthDir);
+		    automatic_tmp_dir = TRUE;
+		    authdir_is_tmp_dir = TRUE;
 	    }
     } else {
 	    authdir = g_strdup (homedir);
     }
 
+    if (d->local_auths != NULL) {
+	    gdm_auth_free_auth_list (d->local_auths);
+	    d->local_auths = NULL;
+    }
+
+    d->local_auths = get_local_auths (d);
+
+    if (d->local_auths == NULL) {
+	    gdm_error ("Can't make cookies");
+	    return FALSE;
+    }
+
 try_user_add_again:
 
     locked = FALSE;
@@ -445,15 +541,34 @@ try_user_add_again:
 	authdir == NULL ||
 	! gdm_file_check ("gdm_auth_user_add", user, authdir, GdmUserAuthFile, 
 			  TRUE, FALSE, GdmUserMaxFile, GdmRelaxPerms) ||
-	! try_open_append (d->userauth)) {
+	! try_open_append (d->userauth) ||
 
-	/* No go. Let's create a fallback file in GdmUserAuthFB (/tmp) */
+	/* try opening as root, if we can't open as root,
+	   then this is a NFS mounted directory with root squashing,
+	   and we don't want to write cookies over NFS */
+	! try_open_read_as_root (d->userauth)) {
+
+	/* No go. Let's create a fallback file in GdmUserAuthFB (/tmp)
+	 * or perhaps GdmUserAuth directory (usually would be /tmp) */
 	d->authfb = TRUE;
 	g_free (d->userauth);
-	d->userauth = g_build_filename (GdmUserAuthFB, ".gdmXXXXXX", NULL);
+	if (authdir_is_tmp_dir && authdir != NULL)
+		d->userauth = g_build_filename (authdir, ".gdmXXXXXX", NULL);
+	else
+		d->userauth = g_build_filename (GdmUserAuthFB, ".gdmXXXXXX", NULL);
 	authfd = g_mkstemp (d->userauth);
 
-	if (authfd == -1) {
+	if G_UNLIKELY (authfd < 0 && authdir_is_tmp_dir) {
+	    g_free (d->userauth);
+	    d->userauth = NULL;
+
+	    umask (022);
+
+	    authdir_is_tmp_dir = FALSE;
+	    goto try_user_add_again;
+	}
+
+	if G_UNLIKELY (authfd < 0) {
 	    gdm_error (_("%s: Could not open cookie file %s"),
 		       "gdm_auth_user_add",
 		       d->userauth);
@@ -472,7 +587,7 @@ try_user_add_again:
 	d->authfb = FALSE;
 
 	/* FIXME: Better implement my own locking. The libXau one is not kosher */
-	if (XauLockAuth (d->userauth, 3, 3, 0) != LOCK_SUCCESS) {
+	if G_UNLIKELY (XauLockAuth (d->userauth, 3, 3, 0) != LOCK_SUCCESS) {
 	    gdm_error (_("%s: Could not lock cookie file %s"),
 		       "gdm_auth_user_add",
 		       d->userauth);
@@ -490,7 +605,7 @@ try_user_add_again:
 	af = gdm_safe_fopen_ap (d->userauth);
     }
 
-    if (!af) {
+    if G_UNLIKELY (af == NULL) {
 	/* Really no need to clean up here - this process is a goner anyway */
 	gdm_error (_("%s: Could not open cookie file %s"),
 		   "gdm_auth_user_add",
@@ -516,13 +631,13 @@ try_user_add_again:
 
     /* If not a fallback file, nuke any existing cookies for this display */
     if (! d->authfb)
-	gdm_auth_purge (d, af);
+	af = gdm_auth_purge (d, af, FALSE /* remove when empty */);
 
     /* Append the authlist for this display to the cookie file */
-    auths = d->auths;
+    auths = d->local_auths;
 
     while (auths) {
-	if ( ! XauWriteAuth (af, auths->data)) {
+	if G_UNLIKELY ( ! XauWriteAuth (af, auths->data)) {
 		gdm_error (_("%s: Could not write cookie"),
 			   "gdm_auth_user_add");
 
@@ -543,7 +658,22 @@ try_user_add_again:
 	auths = auths->next;
     }
 
-    fclose (af);
+    if G_UNLIKELY (fclose (af) < 0) {
+	    gdm_error (_("%s: Could not write cookie"),
+		       "gdm_auth_user_add");
+
+	    if ( ! d->authfb) {
+		    if (locked)
+			    XauUnlockAuth (d->userauth);
+		    g_free (d->userauth);
+		    d->userauth = NULL;
+		    automatic_tmp_dir = TRUE;
+		    goto try_user_add_again;
+	    }
+
+	    ret = FALSE;
+    }
+
     if (locked)
 	    XauUnlockAuth (d->userauth);
 
@@ -572,7 +702,7 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
     gchar *authfile;
     gchar *authdir;
 
-    if (!d || !d->userauth)
+    if G_UNLIKELY (!d || !d->userauth)
 	return;
 
     gdm_debug ("gdm_auth_user_remove: Removing cookie from %s (%d)", d->userauth, d->authfb);
@@ -586,6 +716,13 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
 	return;
     }
 
+    /* if the file doesn't exist, oh well, just ignore this then */
+    if G_UNLIKELY (access (d->userauth, F_OK) != 0) {
+	g_free (d->userauth);
+	d->userauth = NULL;
+	return;
+    }
+
     authfile = g_path_get_basename (d->userauth);
     authdir = g_path_get_dirname (d->userauth);
 
@@ -593,7 +730,7 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
      * decided to concatenate something like his entire MP3 collection
      * to it. So we better play it safe... */
 
-    if ( ! gdm_file_check ("gdm_auth_user_remove", user, authdir, authfile, 
+    if G_UNLIKELY ( ! gdm_file_check ("gdm_auth_user_remove", user, authdir, authfile, 
 			   TRUE, FALSE, GdmUserMaxFile, GdmRelaxPerms)) {
 	    g_free (authdir);
 	    g_free (authfile);
@@ -608,7 +745,7 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
     g_free (authfile);
 
     /* Lock user's cookie jar and open it for writing */
-    if (XauLockAuth (d->userauth, 3, 3, 0) != LOCK_SUCCESS) {
+    if G_UNLIKELY (XauLockAuth (d->userauth, 3, 3, 0) != LOCK_SUCCESS) {
 	g_free (d->userauth);
 	d->userauth = NULL;
 	return;
@@ -616,7 +753,7 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
 
     af = gdm_safe_fopen_ap (d->userauth);
 
-    if (!af) {
+    if G_UNLIKELY (af == NULL) {
 	XauUnlockAuth (d->userauth);
 
 	gdm_error (_("%s: Cannot safely open %s"),
@@ -630,33 +767,67 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
     }
 
     /* Purge entries for this display from the cookie jar */
-    gdm_auth_purge (d, af);
+    af = gdm_auth_purge (d, af, TRUE /* remove when empty */);
 
     /* Close the file and unlock it */
-    fclose (af);
+    if (af != NULL)
+	    /* FIXME: what about out of diskspace errors on errors close */
+	    fclose (af);
+
     XauUnlockAuth (d->userauth);
 
     g_free (d->userauth);
     d->userauth = NULL;
 }
 
+static gboolean
+memory_same (const char *sa, int lena, const char *sb, int lenb)
+{
+	if (lena == lenb) {
+		if (lena == 0)
+			return TRUE;
+		/* sanity */
+		if G_UNLIKELY (sa == NULL || sb == NULL)
+			return FALSE;
+		return memcmp (sa, sb, lena) == 0;
+	} else {
+		return FALSE;
+	}
+}
+
+static gboolean
+auth_same_except_data (Xauth *xa, Xauth *xb)
+{
+	if (xa->family == xb->family &&
+	    memory_same (xa->number, xa->number_length,
+			 xb->number, xb->number_length) &&
+	    memory_same (xa->name, xa->name_length,
+			 xb->name, xb->name_length) &&
+	    memory_same (xa->address, xa->address_length,
+			 xb->address, xb->address_length))
+		return TRUE;
+	else
+		return FALSE;
+}
+
 
 /**
  * gdm_auth_purge:
  * @d: Pointer to a GdmDisplay struct
  * @af: File handle to a cookie file
+ * @remove_when_empty: remove the file when empty
  * 
  * Remove all cookies referring to this display a cookie file.
  */
 
-static void
-gdm_auth_purge (GdmDisplay *d, FILE *af)
+static FILE *
+gdm_auth_purge (GdmDisplay *d, FILE *af, gboolean remove_when_empty)
 {
     Xauth *xa;
     GSList *keep = NULL, *li;
 
-    if (!d || !af)
-	return;
+    if G_UNLIKELY (!d || !af)
+	return af;
 
     gdm_debug ("gdm_auth_purge: %s", d->name);
 
@@ -667,39 +838,47 @@ gdm_auth_purge (GdmDisplay *d, FILE *af)
      * in the cookie jar... */
 
     while ( (xa = XauReadAuth (af)) != NULL ) {
-	gboolean match = FALSE;
-	GSList *alist = d->auths;
-
-	while (alist) {
-	    Xauth *da = alist->data;
-
-	    if (! memcmp (da->address, xa->address, xa->address_length) &&
-		! memcmp (da->number, xa->number, xa->number_length))
-		match = TRUE;
-
-	    alist = alist->next;
+	GSList *li;
+	/* We look at the current auths, but those may
+	   have different cookies then what is in the file,
+	   so don't compare those, but we wish to purge all
+	   the entries that we'd normally write */
+	for (li = d->local_auths; li != NULL; li = li->next) {
+		Xauth *xb = li->data;
+		if (auth_same_except_data (xa, xb)) {
+			XauDisposeAuth (xa);
+			xa = NULL;
+			break;
+		}
 	}
-
-	if (match)
-	    XauDisposeAuth (xa);
-	else
-	    keep = g_slist_append (keep, xa);
+	if (xa != NULL)
+		keep = g_slist_append (keep, xa);
     }
 
-    /* Rewind the file */
-    af = freopen (d->userauth, "w", af);
+    fclose (af);
+
+    if (remove_when_empty &&
+	keep == NULL) {
+	    IGNORE_EINTR (unlink (d->userauth));
+	    return NULL;
+    }
+
+    af = gdm_safe_fopen_w (d->userauth);
 
     /* Write out remaining entries */
     for (li = keep; li != NULL; li = li->next) {
 	    /* FIXME: is this correct, if we can't open
 	     * this is quite crap isn't it ... */
-	    if (af != NULL)
+	    if G_LIKELY (af != NULL)
 		    XauWriteAuth (af, li->data);
+	            /* FIXME: what about errors? */
 	    XauDisposeAuth (li->data);
 	    li->data = NULL;
     }
 
     g_slist_free (keep);
+
+    return af;
 }
 
 void
@@ -709,4 +888,17 @@ gdm_auth_set_local_auth (GdmDisplay *d)
 			   (char *)d->bcookie, (int) 16);
 }
 
+void
+gdm_auth_free_auth_list (GSList *list)
+{
+	GSList *li;
+
+	for (li = list; li != NULL; li = li->next) {
+		XauDisposeAuth ((Xauth *) li->data);
+		li->data = NULL;
+	}
+
+	g_slist_free (list);
+}
+
 /* EOF */
